package icfp.contest;

import java.awt.*;
import java.util.*;

/**
 * Player's state.
 * 
 * @author: Pablo J. Pedemonte
 */
public final class PlayerState {

	/* Constants for possible actions */
	private final static char NORTH_MOVE = 'N';
	private final static char SOUTH_MOVE = 'S';
	private final static char EAST_MOVE  = 'E';
	private final static char WEST_MOVE  = 'W';
	private final static char PKG_PICK   = 'P';
	private final static char PKG_DROP   = 'D';
	private final static char X_DIR      = 'X';
	private final static char Y_DIR      = 'Y';

	/** Constants for state */
	private final static int PICK_MODE = 0;
	private final static int DROP_MODE = 1;
	private final static int INIT_MODE = 2;
		
	/** Board */
	private Board board;

	/** My configuration */
	private PlayerConfig self;
	
	/** Map robots to positions and UID's */
	private Hashtable pn_players = new Hashtable(25);
	private Hashtable id_players = new Hashtable(25);

	/** Map set of packages to their positions */	
	private Hashtable pn_packages = new Hashtable(25);

	/** Some further state variables */
	private int quantum;
	private int origMoney;
	private int sumWeight;
	private int numPackages;
	private HashSet homeCells;
	private int state = INIT_MODE;
	private NeighbourInfo n_info;
/**
 * PlayerState constructor comment.
 * @param port int
 * @param boardFile java.lang.String
 */
public PlayerState(Board board, PlayerConfig self) {
	this.self = self;
	this.board = board;
	this.addPlayerToBoard(self);
	this.origMoney = self.money;
	this.homeCells = board.findHomeCells();
	this.computeQuantum();
	this.n_info = new NeighbourInfo();
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 2:17:50 PM)
 * @param p icfp.contest.Package
 */
public void addPackageToBoard(Package p) {
	this.addPackageToBoard(self, p);
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 2:17:50 PM)
 * @param p icfp.contest.Package
 */
public void addPackageToBoard(PlayerConfig p, Package pkg) {
	Hashtable t = (Hashtable)pn_packages.get(p.pos);
	if (t == null) {
		t = new Hashtable();
		Point freshPos = new Point(p.pos.x, p.pos.y);
		pn_packages.put(freshPos, t);
	}
	t.put(pkg.uid, pkg);

	/** Mantain statistics about average package weight */
	sumWeight += pkg.weight;
	++numPackages;
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 2:17:50 PM)
 * @param p icfp.contest.Package
 */
public void addPlayerToBoard(PlayerConfig p) {
	id_players.put(p.uid, p);
	pn_players.put(p.pos, p);
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 5:33:53 PM)
 * @param p icfp.contest.PlayerConfig
 * @param pos java.lang.String
 * @param dir char
 */
private void changePosTo(PlayerConfig p, String pos, char dir) {
	
	int n = Integer.parseInt(pos);
	pn_players.remove(p.pos);
	
	if (dir == X_DIR) {
		p.pos.x = n - 1;
	}
	else {
		p.pos.y = n - 1;
	}
	
	pn_players.put(p.pos, p);

}
/**
 * Insert the method's description here.
 * Creation date: (9/1/02 7:43:56 PM)
 * @return int
 * @param anxiety int
 */
private int computeBid(int anxiety) {

	if (self.money == 0) {
		return 0;
	}
	
	int aux = Math.abs(anxiety);
	int bid = quantum * aux;
	if (bid == 0 || bid >= self.money) {
		bid = 1;
	}
	self.money -= bid;

	if (anxiety < 0) {
		bid = -bid;
	}
	
	return bid;
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 6:15:22 PM)
 * @return java.lang.String
 */
public String computeNextAction() {

	// No matter what the state is, check if we can
	// drop a package
	ArrayList pkgsToDrop = this.packagesToDrop();
	if (!pkgsToDrop.isEmpty()) {
		return this.computeBid(4) + " Drop " + this.print(pkgsToDrop);
	}
	
	// No matter what the state is, if we are in
	// a package cell try to pick it. In addition,
	// force to calculate where should we go.
	if (pn_packages.containsKey(self.pos)) {
		ArrayList pkgs = this.packagesToPick();
		if (!pkgs.isEmpty()) {
			return this.computeBid(3) + " Pick " + this.print(pkgs);
		}
	}
	
	// Decide in which state we shall operate
	Point toPick = this.findPackages();
	Point toDrop = this.findDests();
	this.defineState(toPick, toDrop);
	Point whereToGo = this.state == PICK_MODE? toPick : toDrop;
		
	// Paranoid test: if we have no packages to deliver, and
	// no packages ot look for, go to the (say) north.
	char dir = whereToGo != null? 
		board.getDir(self.pos, whereToGo) :
		'N';

	this.board.fillInNeighbourInfo(self.pos, dir, this.pn_players, n_info);
	Advice a = n_info.getAdvice();
		
	return this.computeBid(this.getBidForAdvice(a)) + " Move " + a.getDir();
}
/**
 * Insert the method's description here.
 * Creation date: (9/1/02 10:16:15 PM)
 */
private void computeQuantum() {
	quantum = origMoney / (board.getWidth() * board.getHeight() * 4);
	if (quantum == 0) {
		quantum = 1;
	}
}
/**
 * Define the player's state (PICKER or DROPPER)
 *
 * @return <code>true</code> if the mode changed,
 * <code>false</code> otherwise
 */
private void defineState(Point toPick, Point toDrop) {
	// If we have no packages, pick some!
	if (self.packages.size() == 0) {
		state = PICK_MODE;
		return;
	}

	// We spent more than the half of our budget, start
	// delivering our packages
	if (self.money < this.origMoney / 2) {
		state = DROP_MODE;
		return;
	}

	// If our capacity is less than the average package
	// weight, start dropping packages
	if (self.capacity <= this.getAvgPackageWeight()) {
		state = DROP_MODE;
		return;
	}

	// Otherwise, decide based on distances to pck and drop
	if (toPick == null) {
		state = DROP_MODE;
	}
	else if (toDrop == null) {
		state = PICK_MODE;
	}
	else {
		int d1 = board.getManhattanDist(self.pos, toPick);
		int d2 = board.getManhattanDist(self.pos, toDrop);
		state = d1 < d2? PICK_MODE : DROP_MODE;
	}

	return;
}
/**
 * Insert the method's description here.
 * Creation date: (9/1/02 10:31:02 PM)
 * @return java.awt.Point
 */
private Point findDests() {
	
	int dist   = 0;
	int val    = 0;
	int maxVal = 0;
	Point dest = null;
	Package pkg;

	for (Iterator it = self.packages.values().iterator(); it.hasNext(); ) {
		pkg  = (Package)it.next();
		dist = board.getManhattanDist(self.pos, pkg.pos);
		val  = 100 * pkg.weight / (dist * dist);
		if (val > maxVal) {
			maxVal = val;
			dest = pkg.pos;
		}
	}

	return dest;
}
/**
 * Insert the method's description here.
 * Creation date: (9/1/02 7:14:24 PM)
 * @return java.awt.Point
 */
private Point findPackages() {

	int n;
	int dist;
	double val = 0;
	double maxVal = 0;
	Point p = null;
	Point choice = null;
	int choiceN = 0;
	
	// First, choose the most valuable location for the
	// dropped packages
	for (Iterator it = pn_packages.keySet().iterator(); it.hasNext(); ) {
		p = (Point)it.next();
		n = ((Hashtable)pn_packages.get(p)).size();
		if (n == 0) {
			break;
		}
		dist    = board.getManhattanDist(self.pos, p);		
		val     = dist == 0? Integer.MAX_VALUE : 100 * n / (dist * dist);
		if (val > maxVal) {
			maxVal = val;
			choice = p;
			choiceN = n;
		}
	}

	// Choose the nearest location for home packages
	Point minPoint = null;
	int minDist = Integer.MAX_VALUE;
	for (Iterator it = homeCells.iterator(); it.hasNext(); ) {
		p = (Point)it.next();
		dist = board.getManhattanDist(self.pos, p);
		if (dist < minDist) {
			minDist = dist;
			minPoint = p;
		}
	}

	if (choice == null) {
		return minPoint;
	}

	if (minPoint == null) {
		return choice;
	}
	
	// Heuristic return
	if (choiceN >= id_players.size() / 2 + 1) {
		return choice;
	}
	else {
		int d1 = board.getManhattanDist(self.pos, choice);
		int d2 = board.getManhattanDist(self.pos, minPoint);
		return d1 <= d2? choice : minPoint;
	}
}
/**
 * Insert the method's description here.
 * Creation date: (9/1/02 10:50:39 PM)
 * @return double
 */
private double getAvgPackageWeight() {
	if (numPackages == 0) {
		return 0;
	}

	return sumWeight / numPackages;
}
/**
 * Insert the method's description here.
 * 
 * @return int
 * @param a icfp.contest.Advice
 */
private int getBidForAdvice(Advice a) {
	switch(a.getAction()) {
		case Advice.HIGH_BID: return 5;
		case Advice.NORM_BID: return 2;
		case Advice.LOW_BID : return -5;
		default: return 2;
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 5:20:29 PM)
 * @param p icfp.contest.PlayerConfig
 * @param dir char
 */
private void moveToNewPos(PlayerConfig p, int x_delta, int y_delta) {
	pn_players.remove(p.pos);
	p.pos.x += x_delta;
	p.pos.y += y_delta;
	pn_players.put(p.pos, p);
}
/**
 * Return the set of droppeable packages.
 */
private ArrayList packagesToDrop() {

	Package p;
	ArrayList pkgs = new ArrayList();
	for (Iterator it = self.packages.values().iterator(); it.hasNext(); ) {
		p = (Package)it.next();
		if (p.pos.equals(self.pos)) {
			pkgs.add(p.uid);
		}
	}

	return pkgs;
}
/**
 * We are in a cell containing packages. Try to pick one!
 * Creation date: (9/1/02 7:27:20 PM)
 */
private ArrayList packagesToPick() {
	
	Hashtable t = (Hashtable)pn_packages.get(self.pos);

	Package p = null;
	ArrayList choices = new ArrayList();
	for (Iterator it = t.values().iterator(); it.hasNext(); ) {
		p = (Package)it.next();
		
		if (p.weight > self.capacity || p.pos.equals(self.pos)) {
			continue;
		}

		choices.add(p.uid);
	}

	// Try to remove this point from my set of home cells.
	// It's safe to do so, because my package info was already
	// updated and we know that there are packages here
	homeCells.remove(self.pos);
	
	return choices;
}
/**
 * Insert the method's description here.
 * Creation date: (9/1/02 10:12:12 PM)
 * @return java.lang.String
 * @param a java.util.ArrayList
 */
private String print(ArrayList a) {
	StringBuffer b = new StringBuffer();
	for(int i = 0; i < a.size(); i++) {
		b.append(a.get(i));
		if (i < a.size() - 1) {
			b.append(' ');
		}
	}

	return b.toString();
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 5:39:48 PM)
 * @param packetID java.lang.String
 */
private void processPkgDrop(PlayerConfig p, String uid) {
	Package pkg = p.removePackage(uid);

	// Avoid grabbing again my own dropped packages
	if (!self.equals(p)) {
		this.addPackageToBoard(p, pkg);
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 5:39:48 PM)
 * @param packetID java.lang.String
 */
private void processPkgPick(PlayerConfig p, String uid) {
	Hashtable t = (Hashtable)pn_packages.get(p.pos);
	Package pkg = (Package)t.remove(uid);
	p.addPackage(pkg);
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 2:58:19 PM)
 * @param updateInfo java.lang.String
 */
public void processRobotUpdates(String updateInfo) {

	String uid;
	String item;
	PlayerConfig p = null;
	StringTokenizer t = new StringTokenizer(updateInfo, " \t");
	
	while(t.hasMoreTokens()) {
		
		item = t.nextToken();
		if (item.startsWith("#")) {
			uid = item.substring(1);
			p = (PlayerConfig)id_players.get(uid);
		
			if (p == null) {
				p = new PlayerConfig();
				p.uid = uid;
				this.addPlayerToBoard(p);
			}
		}
		else {
			switch (item.charAt(0)) {
				case NORTH_MOVE: this.moveToNewPos(p, 0, 1) ; break;
				case SOUTH_MOVE: this.moveToNewPos(p, 0, -1); break;
				case EAST_MOVE : this.moveToNewPos(p, 1, 0) ; break;
				case WEST_MOVE : this.moveToNewPos(p, -1, 0); break;
				case PKG_PICK  : this.processPkgPick(p, t.nextToken()); break;
				case PKG_DROP  : this.processPkgDrop(p, t.nextToken()); break;
				
				case X_DIR     : this.changePosTo(p, t.nextToken(), X_DIR); 
				                 break;
				case Y_DIR     : this.changePosTo(p, t.nextToken(), Y_DIR); 
				                 break;
			}
		}
	}
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 2:17:50 PM)
 * @param p icfp.contest.Package
 */
public void removePackageFromBoard(Package p) {
	Hashtable t = (Hashtable)pn_packages.get(p.pos);
	t.remove(p.uid);
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 2:17:50 PM)
 * @param p icfp.contest.Package
 */
public void removePlayerFromBoard(PlayerConfig p) {
	id_players.remove(p.uid);
	pn_players.remove(p.pos);
}
/**
 * Insert the method's description here.
 * Creation date: (8/31/02 6:12:11 PM)
 * @return java.lang.String
 */
public String toString() {
	StringBuffer b = new StringBuffer();
	b.append("Self: ").append(self).append('\n');
	b.append("Board:\n").append(board).append('\n');
	b.append("Packages:\n").append(pn_packages).append('\n');
	b.append("Players (by UID):\n").append(id_players).append('\n');
	b.append("Players (by POS):\n").append(pn_players);
	return b.toString();
}
}
